/****************************************************************************
**
** Qtitan    Library by Developer Machines (Advanced Grid for Qt)
** 
** Copyright (c) 2009-2011 Developer Machines (http://www.devmachines.com)
**           ALL RIGHTS RESERVED
** 
**  The entire contents of this file is protected by copyright law and
**  international treaties. Unauthorized reproduction, reverse-engineering
**  and distribution of all or any portion of the code contained in this
**  file is strictly prohibited and may result in severe civil and 
**  criminal penalties and will be prosecuted to the maximum extent 
**  possible under the law.
**
**  RESTRICTIONS
**
**  THE SOURCE CODE CONTAINED WITHIN THIS FILE AND ALL RELATED
**  FILES OR ANY PORTION OF ITS CONTENTS SHALL AT NO TIME BE
**  COPIED, TRANSFERRED, SOLD, DISTRIBUTED, OR OTHERWISE MADE
**  AVAILABLE TO OTHER INDIVIDUALS WITHOUT WRITTEN CONSENT
**  AND PERMISSION FROM DEVELOPER MACHINES
**
**  CONSULT THE END USER LICENSE AGREEMENT FOR INFORMATION ON
**  ADDITIONAL RESTRICTIONS.
**
****************************************************************************/

#ifndef QTN_MODELCONTROLLER_H
#define QTN_MODELCONTROLLER_H

#include <QVector>
#include <QSet>
#include <QAbstractItemModel>

#include "QtitanDef.h"

namespace Qtitan
{
    class GridModelController;
    class GridRow;
    class GridCell;
    class GridPersistentRowIndex;
    class GridSummary;
    class ModelDataBinding;
    class RelationEventArgs;
    class GridFilter;

    enum SortOrder
    {
        SortNone,
        SortAscending,
        SortDescending
    };

    struct GridSortColumnRule
    {
        int column;
        SortOrder order;
        bool caseSensitivity;
        bool localeAware;
    };

    class GridSortColumnRules
    {
    public:
        GridSortColumnRules();
        void clear();
        void clearSorting();
        void clearGrouping();
        void setSortRule(int column, int index, SortOrder order, bool caseSensitivity = false, bool localeAware = false);
        void setGroupRule(int column, int index, SortOrder order, bool caseSensitivity = false, bool localeAware = false);
        int count() const;
        const GridSortColumnRule& rule(int index) const;
        int groupRulesCount() const;
        const GridSortColumnRule& groupRule(int index) const;
        const GridSortColumnRule* groupRuleByColumn(int column) const;
        int sortRulesCount() const;
        const GridSortColumnRule& sortRule(int index) const;
        const GridSortColumnRule* sortRuleByColumn(int column) const;
    protected:
    private:
        QList<GridSortColumnRule> m_sortRulesList;
        QList<GridSortColumnRule> m_groupRulesList;
    };

    class ModelGroup
    {
    public:
        int column;
        int depth;
        int row;
        int index;       // Index of group structure in the controller group list.
        int parentIndex; // Index of parent group structure in the controller group list.
        bool last;
        int firstRecord;
        int lastRecord;
        bool expanded;
        QVariant groupValue; // Value on which group is formed.
    };

    class GridModelValueCache
    {
    public:
        GridModelValueCache(GridModelController* controller);
        virtual ~GridModelValueCache();
        void setDimension(int columnCount, int rowCount);
        void insertRow(int row);
        void removeRow(int row);
        bool getValue(int column, int row, QVariant& retval, Qt::ItemDataRole role = Qt::DisplayRole);
        bool setValue(int column, int row, const QVariant& value, Qt::ItemDataRole role = Qt::EditRole);
        void clear();
        Qt::ItemDataRole cachedRole() const;
        void setCachedRole(Qt::ItemDataRole role);
    protected:
        typedef QVector<QVariant> ColumnValues;
        typedef QVector<ColumnValues> TableValues;
    private:
        GridModelController* m_controller;
        QVector<QVariant> m_values;
        Qt::ItemDataRole m_cachedRole;
    };

    struct RecordMapKey
    {
        RecordMapKey(const QVariant& v, int c)
        {
            value = v;
            column = c;
        };
        QVariant value;
        int column;
    };

    inline uint qHash(const Qtitan::RecordMapKey &key)
    {
        return qHash(key.value.toString()) + key.column;
    }

    inline bool operator==(const RecordMapKey& first, const RecordMapKey& second)
    {
        return first.value == second.value && first.column == second.column;
    }

    typedef QHash<RecordMapKey, bool> GroupingSavedState;

    class GridPersistentRowIndex
    {
    public:
        friend class GridCellSelectionRange;
        friend class GridModelController;
        GridPersistentRowIndex(GridModelController* controller);
        GridPersistentRowIndex(const GridPersistentRowIndex& index);
        virtual ~GridPersistentRowIndex();
        void reset();
        void set(GridRow* gridRow);
        bool equal(GridRow* gridRow) const;
        void setCellIndex(int index);
        int cellIndex() const;
        bool isCellSet(int index) const;
        GridRow* row() const;
        int groupColumn;
        int record;
        int modelRow;
        GridPersistentRowIndex& operator=(const GridPersistentRowIndex& index);
    private:
        GridModelController* m_controller;
        int m_cellIndex;
    };

    enum CacheUsage
    {
        NoCache,
        UseCache,
        UseCacheImmediate
    };

    enum SummaryType
    {
        SummaryNone = 0,
        SummarySum = 0x0001,
        SummaryMin = 0x0002,
        SummaryMax = 0x0004,
        SummaryCount = 0x0008,
        SummaryAvg = 0x0010
    };
    Q_DECLARE_FLAGS(SummaryTypes, SummaryType)


    class NotifyHelper
    {
    public:
        virtual void raiseQueryRelation(RelationEventArgs* args) = 0;
        virtual void raiseGroupExpanding(ModelGroup* group, bool& allow) = 0;
        virtual void raiseGroupExpanded(ModelGroup* group) = 0;
        virtual void raiseGroupCollapsing(ModelGroup* group, bool& allow) = 0;
        virtual void raiseGroupCollapsed(ModelGroup* group) = 0;
    };

    class QTITAN_EXPORT ModelDataBinding: public QObject
    {
        Q_OBJECT
    public:
        enum RelationMode
        {
            NoRelation = 0,
            Relation = 1,
            MasterDetailsRelation = 2
        };

    public:
        ModelDataBinding(GridModelController* modelController, int column);
        virtual ~ModelDataBinding();
        int column() const;
        void setColumn(int column);
        QString columnName() const;
        QVariant defaultValue(Qt::ItemDataRole role = Qt::EditRole) const;
        void setDefaultValue(const QVariant& value, Qt::ItemDataRole role = Qt::EditRole);
        Qt::ItemDataRole displayRole() const;
        void setDisplayRole(Qt::ItemDataRole role = Qt::DisplayRole);
        Qt::ItemDataRole editRole() const;
        void setEditRole(Qt::ItemDataRole role = Qt::EditRole);
        Qt::ItemDataRole summaryRole() const;
        void setSummaryRole(Qt::ItemDataRole role = Qt::DisplayRole);
        SummaryTypes supportedSummaryTypes() const;
        void setSupportedSummaryTypes(SummaryTypes types);
        RelationMode relationMode() const;
        void setRelationMode(RelationMode mode);
        QAbstractItemModel* relationModel();
        void setRelationModel(QAbstractItemModel* model, QModelIndex parentIndex = QModelIndex());
        const QModelIndex& relationParentIndex() const;
        void setForeignKey(int foreignColumn, Qt::ItemDataRole foreignRole = Qt::DisplayRole, Qt::ItemDataRole masterRole = Qt::EditRole);
        void setForeignKey(const QString& foreignColumnName, Qt::ItemDataRole foreignRole = Qt::DisplayRole, Qt::ItemDataRole masterRole = Qt::EditRole);
        int foreignKeyColumn() const;
        QString foreignKeyColumnName() const;
        Qt::ItemDataRole foreignKeyRole() const;
        Qt::ItemDataRole foreignKeyMasterRole() const;
        void setRoleMapping(Qt::ItemDataRole role, const QString& foreignColumnName, Qt::ItemDataRole foreignRole = Qt::DisplayRole);
        void setRoleMapping(Qt::ItemDataRole role, int foreignColumn, Qt::ItemDataRole foreignRole = Qt::DisplayRole);
        int getForeignColumnMapping(Qt::ItemDataRole role) const;
        QString getForeignColumnNameMapping(Qt::ItemDataRole role) const;
        Qt::ItemDataRole getForeignRoleMapping(Qt::ItemDataRole role) const;
    public:
        bool getValue(GridRow* row, Qt::ItemDataRole role, QVariant& retval) const;
        bool setValue(GridRow* row, Qt::ItemDataRole role, const QVariant& value);
        QModelIndex findRelationModelIndex(const QVariant& foreignKeyValue) const;
    private:
        int columnByColumnName(const QString& columnName) const;
        bool isRelationModelColumnRoleSupported(Qt::ItemDataRole role) const;
        QModelIndex getRelationModelIndex(GridRow* row, Qt::ItemDataRole role, Qt::ItemDataRole& foreignRole) const;
        GridModelController* m_modelController;
        int m_column;
        Qt::ItemDataRole m_displayRole;
        Qt::ItemDataRole m_editRole;
        Qt::ItemDataRole m_summaryRole;
        SummaryTypes m_supportedSummaryTypes;
        RelationMode m_relationMode;
        QAbstractItemModel* m_relationModel;
        QModelIndex m_relationParentIndex;
        int m_foreignKeyColumn;
        Qt::ItemDataRole m_foreignKeyMasterRole;
        Qt::ItemDataRole m_foreignKeyRole;
        struct IndexColumn
        {
            IndexColumn()
                : role(Qt::UserRole)
            {
            }
            IndexColumn(int c, Qt::ItemDataRole r)
                : column(c), role(r)
            {
            }
            int column;
            Qt::ItemDataRole role;
        };
        typedef QHash<Qt::ItemDataRole, IndexColumn> ForeignColumnRoleMap;
        ForeignColumnRoleMap m_foreignColumnRoleMap;
    };

    class QTITAN_EXPORT RelationEventArgs: public BaseEventArgs
    {
    public:
        RelationEventArgs(GridModelController* controller, ModelDataBinding* binding);
        ModelDataBinding* dataBinding() const;
    private:
        GridModelController* m_controller;
        ModelDataBinding* m_dataBinding;
    };

    class QTITAN_EXPORT GridCell
    {
    public:
        GridCell();
        GridCell(GridRow* r, int ci);
        GridRow* row() const;
        int rowIndex() const;
        int columnIndex() const;
    private:
        GridRow* m_row;
        int m_index;
    };

    inline uint qHash(const GridCell& cell)
    { 
        return uint((cell.rowIndex() << 4) + cell.columnIndex());
    }
    inline bool operator==(const GridCell& first, const GridCell& second)
    {
        return first.row() == second.row() && first.columnIndex() == second.columnIndex();
    }


    class QTITAN_EXPORT GridRow
    {
    public:
        enum ModelRowType
        {
            DataRow,
            GroupRow
        };
    public:
        friend class GridPersistentRowIndex;
        friend class GridModelController;
        GridRow();
        ModelRowType type() const;
        int depth() const;
        int rowIndex() const;
        int modelRowIndex() const;
        GridRow* parentGroupRow();
        inline ModelGroup* modelGroup() const { return m_group; }
        bool last() const;
        bool expanded() const;
        void setExpanded(bool expanded);
        int groupColumn() const;
        GridCell cell(int index) const;
        bool groupValue(QVariant& retval) const;
        GridModelController* controller() const;
    public:
        static bool compareByYpos(GridRow* modelRow, const int& ypos);
        static bool compareByRecord(GridRow* modelRow, const int& record);
    protected:
        void initialize(GridModelController* controller, ModelRowType type, ModelGroup* group, int rowIndex);
        void reset();
        bool isInitialized() const;
        int record() const;
        int m_ypos;
    private:
        GridModelController* m_controller;
        ModelRowType m_type;
        int m_rowIndex;
        ModelGroup* m_group;
        bool m_initialized;
    };

    class QTITAN_EXPORT GridCellSelectionRange
    {
    public:
        GridCellSelectionRange(GridModelController* controller, const GridCell& topLeft, const GridCell& bottomRight);
        GridCell topLeft() const;
        GridCell bottomRight() const;
        bool isCellSelected(const GridCell& cell) const;
        bool isRowSelected(const GridRow* row) const;
        bool isValid() const;
        bool contains(const GridCellSelectionRange& range) const;
        GridCellSelectionRange intersect(const GridCellSelectionRange& range) const;
    private:
        GridPersistentRowIndex m_persistTopLeft;
        GridPersistentRowIndex m_persistBottomRight;
    };

    typedef QList<GridCellSelectionRange> SelectionRanges;

    class QTITAN_EXPORT GridSelectionIterator
    {
    public:
        GridSelectionIterator(GridModelController* controller);
        void reset();
        bool next();
        bool end();
        GridRow* row() const;
        int column() const;
    private:
        GridModelController* m_controller;
        QList<GridCell>::const_iterator current;
        QList<GridCell> m_select;
    };

    class QTITAN_EXPORT GridSelection
    {
    public:
        friend class GridModelController;
        GridSelection(GridModelController* controller);
        int getSelectedRowCount() const;
        int getSelectedRowIndex(int i) const;
        int getSelectedColumnCount(GridRow* row) const;
        int getSelectedColumnIndex(GridRow* row, int i) const;
    protected:
        void update();
    private:
        GridModelController* m_controller;
        QVector<const GridPersistentRowIndex *> m_selected;
    };

    class QTITAN_EXPORT GridModelController: public QObject
    {
        Q_OBJECT
    public:
        friend class ModelDataBinding;
        friend class GridRow;
        friend class GridModelValueCache;
        friend class GridSummary;
        friend class GridSummaryCalculator;
        friend class GridSelection;
        friend class GridFilter;
        friend class GridSelectionIterator;
        friend class GridPersistentRowIndex;
        typedef QVector<QVariant> ColumnValueCache;
        typedef QVector<ColumnValueCache> ModelValueCache;
        typedef QVector<ModelGroup> ModelGroupList;
        GridModelController(NotifyHelper* helper);
        virtual ~GridModelController();
        void setModel(QAbstractItemModel* model, QModelIndex parent = QModelIndex());
        QAbstractItemModel* model() const;
        const QModelIndex& parentIndex() const;
        void setSortOrder(int column, int sortIndex, SortOrder order, bool caseSensitivity = false, bool localeAware = false);
        SortOrder getSortOrder(int column) const;
        void setGroupOrder(int column, int groupIndex, SortOrder order, bool caseSensitivity = false, bool localeAware = false);
        SortOrder getGroupOrder(int column) const;
        void clearSorting();
        void clearGrouping();
        bool isGrouping() const;
        int groupingDepth() const;
        int getModelColumnCount() const;
        int getGridRowCount() const;
        GridRow* getGridRow(int index);
        GridRow* getGridRowByModelRow(int row);
        int modelGroupCount() const;
        ModelGroup* modelGroup(int index);
        ModelGroupList& groups();
        void expandAllGroup();
        void collapseAllGroup();
        bool expandGroup(ModelGroup* group, bool disableChanged = false);
        bool collapseGroup(ModelGroup* group, bool disableChanged = false);
        void refresh();
        GridSummary* summary() const;
        GridFilter* filter() const;
        bool getModelValue(int column, int modelRow, Qt::ItemDataRole role, QVariant& retval) const;
        bool setModelValue(int column, int modelRow, Qt::ItemDataRole role, const QVariant& retval);
        int ensureVisible(int modelRow);
        int focusedRowIndex();
        void setFocusedRowIndex(int row);
        GridRow* focusedRow();
        void setFocusedRow(GridRow* row);
        bool isRowFocused(GridRow* row);
        void setFocusedCell(int index);
        int focusedCell() const;
        bool isCellFocused(GridRow* row, int cellIndex) const;
        GridSelection* selection() const;
        GridSelectionIterator* selectionIterator() const;
        void beginUpdateData();
        void endUpdateData();
        bool isDataUpdating() const;
        CacheUsage cacheUsage() const;
        void setCacheUsage(CacheUsage usage);
        bool insertModelRow(int row);
        void beginAddRow(int row);
        void endAddRow();
        QVariant defaultValue(int column, Qt::ItemDataRole role = Qt::EditRole) const;
        void setDefaultValue(int column, const QVariant& value, Qt::ItemDataRole role = Qt::EditRole);
    public:
        void beginSelect();
        void endSelect();
        void clearSelect();
        //void selectAll();
        void selectRanges(SelectionRanges& ranges);
        void selectCurrent();
        bool isRowSelected(const GridRow* row) const;
        bool isCellSelected(const GridCell& cell) const;
        const GridPersistentRowIndex* selectedRowByRowIndex(int rowIndex);
        int getRowIndexBySelectedRowIndex(const GridPersistentRowIndex* row);
    Q_SIGNALS:
        void changed();
        void selectChanged();
    protected:
        void controllerChanged();
        void clearModelGroups();
        void clearPersistRows();
        void createModelGroups(bool restoreState);
        void resizeModelRows(int size);
        ModelGroup* appendModelGroup();
        void insertDataRow(int modelRowIndex);
        void removeDataRow(int modelRowIndex);
        void fill(bool sort);
        void fillRows();
        ModelGroup* findGroup(int row, int& groupIndex);
        int findRowIndexByRecord(int record);
        int findRecordByModelRow(int modelRow);
        static int compareModelRows(GridModelController* controller, 
                                    const int& firstRow, const int& secondRow);
        void saveState();
        void restoreSelectState();
        void clearGroupingState();
    protected slots:
        void modelLayoutChanged();
        void modelDataChanged(const QModelIndex& topLeft, const QModelIndex& bottomRight);
        void modelRowsAboutToBeInserted(const QModelIndex& parent, int start, int end);
        void modelRowsAboutToBeRemoved(const QModelIndex& parent, int start, int end);
        void modelRowsInserted(const QModelIndex& parent, int start, int end);
        void modelRowsRemoved(const QModelIndex& parent, int start, int end);
    private:
        QAbstractItemModel* m_model;
        QModelIndex m_parentIndex;
        QVector<int> m_recordsMap;
        ModelGroupList m_groups;
        GridSortColumnRules m_sortRules;
        GridModelValueCache* m_modelCache;
        QVector<GridRow> m_modelRows;
        GroupingSavedState m_groupingSavedState;
        GridSummary* m_summary;
        GridFilter* m_filter;
        GridPersistentRowIndex* m_focusedRowIndex;
        SelectionRanges m_selectionRanges;
        SelectionRanges m_currentSelectionRanges;
        int m_currentSelectionAction;
        int m_selectStack;
        GridSelection* m_selection;
        GridSelectionIterator* m_selectionIterator;
        QSet<int> m_insertModelRows;
        QSet<int> m_removeModelRows;
        int m_updateDataStack;
        CacheUsage m_cacheUsage;
        QSet<GridPersistentRowIndex* > m_persistRows;
        int m_addrow;
        QMap<int, QVariant> m_defValues;
        NotifyHelper* m_helper;
    };

    class SelectAutoUpdater
    {
    public:
        SelectAutoUpdater(GridModelController* controller)
            : m_controller(controller)
        {
            m_controller->beginSelect();
        }

        ~SelectAutoUpdater()
        {
            m_controller->endSelect();
        }
    private:
        GridModelController* m_controller;
    };

    class ModelControllerAutoUpdater
    {
    public:
        ModelControllerAutoUpdater(GridModelController* controller)
            : m_controller(controller)
        {
            m_controller->beginUpdateData();
        }

        ~ModelControllerAutoUpdater()
        {
            m_controller->endUpdateData();
        }
    private:
        GridModelController* m_controller;
    };

    extern int find_modelRowIndex_by_view(GridModelController* controller, int ypos);


    struct SummaryValue
    {
         SummaryType type;
        Qt::ItemDataRole role;
        int count;
        QVariant value;
        SummaryValue() : type(SummaryNone), role(Qt::DisplayRole), count(0) {}
    };

    class SummaryTextEventArgs: public BaseEventArgs
    {
    public:
        SummaryTextEventArgs(ModelDataBinding* binding, const QString& text);
        ModelDataBinding* dataBinding() const;
        const QString& text() const;
        void setText(const QString& t);
    private:
        ModelDataBinding* m_binding;
        QString m_text;
    };

    class SummaryGroupTextEventArgs: public SummaryTextEventArgs
    {
    public:
        SummaryGroupTextEventArgs(ModelDataBinding* binding, int modelGroupColumn, const QString& text);
        int modelGroupColumn() const;
    private:
        int m_modelGroupColumn;
    };

    typedef QHash<int, SummaryValue> ColumnSummaryValues;

    class GridSummarySetting
    {
    public:
        GridSummarySetting(int column, Qt::ItemDataRole role,  SummaryType type);
        int column() const;
        Qt::ItemDataRole role() const;
         SummaryType type() const;
    private:
        int m_column;
        Qt::ItemDataRole m_role;
         SummaryType m_type;
    };

    class GridSummaryGroupSetting: public GridSummarySetting
    {
    public:
        enum ShowingPlace
        {
            FooterPlace,
            GroupPlace
        };
        GridSummaryGroupSetting(int groupColumn, int column, Qt::ItemDataRole role,  SummaryType type, ShowingPlace place = FooterPlace);
        int groupColumn() const;
        bool sorted() const;
        void setSorted(bool sorted);
        ShowingPlace showingPlace() const;
        void setShowingPlace(ShowingPlace place);
    private:
        int m_groupColumn;
        bool m_sorted;
        ShowingPlace m_showingPlace;
    };

    class QTITAN_EXPORT GridSummary: public QObject
    {
        Q_OBJECT
    public:
        GridSummary(GridModelController* controller);
        void setFooterSetting(int column, Qt::ItemDataRole role,  SummaryType type);
        void setDefaultGroupSetting(int column, Qt::ItemDataRole role,  SummaryType type);
        void setGroupSetting(int groupColumn, int column, Qt::ItemDataRole role,  SummaryType type);

        const GridSummarySetting& footerSetting(int index) const;
        int footerSettingsCount() const;

        const GridSummaryGroupSetting& defaultGroupSetting(int index) const;
        int defaultGroupSettingsCount() const;

        const GridSummaryGroupSetting& groupSetting(int groupColumn, int index) const;
        int groupSettingsCount(int groupColumn) const;
        void getGroupSummarySettings(int groupColumn, QVector<const GridSummaryGroupSetting *>& settings) const;
        void calculate();
        bool do_calculate_summary(int groupColumn);
        const SummaryValue& getFooterSummaryValue(int column) const;
        QString getFooterSummaryDisplayText(int column) const;
        const SummaryValue& getGroupSummaryValue(ModelGroup* group, int column) const;
        QString getGroupSummaryDisplayText(ModelGroup* group, int column) const;
        void beginUpdate();
        void endUpdate();
        bool supportSummaryType(int column,  SummaryType type) const;
    Q_SIGNALS:
        void calculated();
    private:
        void createDisplayText(const SummaryValue& value, bool isFooterSummary, QString& retval) const;
        GridModelController* m_controller;
        QMap<int, GridSummarySetting> m_footerSummaryMap;
        QMap<int, QMap<int, GridSummaryGroupSetting> > m_groupSummaryMap;
        QVector<SummaryValue> m_footerResults;
        QVector<SummaryValue> m_groupResults;
        int m_updateCount;
    };

    class GridSummaryCalculator
    {
    public:
        GridSummaryCalculator(GridSummary* summary, GridModelController* controller);
        void start(int column, Qt::ItemDataRole role,  SummaryType type);
        SummaryValue stop(int column);
        void calculate(int firstRecord, int secondRecord);
        void clear();
    private:
        void do_calculate(int firstRecord, int secondRecord, int column, SummaryValue& values);
    private:
        void startCalculation(ColumnSummaryValues& hash, int column, Qt::ItemDataRole role,  SummaryType type);
        SummaryValue stopCalculation(ColumnSummaryValues& hash, int column);
        ColumnSummaryValues m_columnValuesHash;
        GridSummary* m_summary;
        GridModelController* m_controller;
    };


    class GridFilterItem
    {
    public:
        GridFilterItem(int column, Qt::ItemDataRole role, const QVariant& value)
            : m_column(column), m_role(role), m_value(value) {}
        int column() const { return m_column; }
        Qt::ItemDataRole role() const { return m_role; }
        const QVariant& value() const { return m_value; }
    private:
        int m_column;
        Qt::ItemDataRole m_role;
        QVariant m_value;
    };

    class GridFilter
    {
    public:
        GridFilter(GridModelController* controller);
        GridFilterItem* addFilterItem(int column, Qt::ItemDataRole role, const QVariant& value);
        void removeItem(GridFilterItem* item);
        void clear();
        bool validateRow(int modelRow);
    private:
        GridModelController* m_controller;
        QList<GridFilterItem *> m_items;
    };
}; //namespace Qtitan

Q_DECLARE_OPERATORS_FOR_FLAGS(Qtitan::SummaryTypes)

#endif //QTN_MODELCONTROLLER_H

